# genkeys.py
# Generate Enigma I (or M3 or ...) keys list

import sys
import random
import getopt
import string

ALPHA = string.ascii_uppercase
WHEELS=['I   ','II  ','III ','IV  ','V   ','VI  ','VII ','VIII']
FIELDS={ 'D':'Tag', 'W':'Walzenlage    ', 'R':'Ringstellung', 'E': ' ',
 'G':'Grundstellung', 'S':'Steckerverbindungen     ', 'K':'Kenngruppen', 
 'g':'Grund.', 'r':'Ringst.', 'e': ' ', 'k':'Kennzeichnung' }
# ====================================================
def genAllKenn():
    for l1 in ALPHA:    
        for l2 in ALPHA:
            for l3 in ALPHA:
                aKenn = l1 + l2 + l3
                allKenn.append( aKenn )
# ====================================================
def genKzeich():
    for day in range(31):
        borne_gauche = 0 + 10 * random.randint(1,8)
        borne_droite = 980 - borne_gauche
        entre = ( (borne_droite - borne_gauche ) // ( NB_OF_NETS * 10 ) ) * 10
        tb = list()
        debut = borne_gauche + 10 * random.randint(0,1)
        if NB_OF_NETS == 1: 
            entre //= 2
            debut += 10 * random.randint(0,40)
        for i in range( NB_OF_NETS ):
            fin = debut + entre - 10 * random.randint(0,1)
            if fin == 1000: fin = 990
            ch = "%03d-%03d" % (debut , fin )
            debut = fin + 10 
            tb.append( ch )
        random.shuffle( tb )
        allKzeich.append( tb )

# ====================================================
def walz( n ):
    walzenlage = ""
    theWalz = list( range(n) )
    random.shuffle( theWalz )
    for i in range(3):
        x = theWalz[0] 
        walzenlage += WHEELS[ x ] + " "
        theWalz.remove( x )
    return walzenlage
# ====================================================
def trig():
    trigram = ""
    for i in range(3):
        x = random.randint(0,25)
        trigram += ALPHA[x]
    return trigram
# ====================================================
def ring():
    ringstellung = ""
    for i in range(3):
        x = random.randint(0,25)
        if RING_NUMB:
            if ZERO:
                ringstellung += "%02d " % (x+1)
            else:
                ringstellung += "%2d " % (x+1)
        else:
            ringstellung += ALPHA[ x ] + " "
    return ringstellung
# ====================================================
def grund():
    grundstellung = ""
    for i in range(3):
        x = random.randint(0,25)
        if GRUND_NUMB:
            if ZERO:
                grundstellung += "%02d " % (x+1)
            else:
                grundstellung += "%2d " % (x+1)
        else:
            grundstellung += ALPHA[ x ] + " "
    return grundstellung
# ====================================================
def steckers():
    tb = list( range(26) )
    nbplugs = 0
    steckerverbindungen = ""
    nb_of_plugs = NB_OF_PLUGS
    if MAX_PLUGS != MIN_PLUGS:
        nb_of_plugs = random.randint( MIN_PLUGS, MAX_PLUGS )
    while nbplugs < nb_of_plugs :
        x = tb[ random.randint( 0, len(tb)-1 )]
        y = tb[ random.randint( 0, len(tb)-1 )]
        if x == y: continue
        if CSKO:
            if abs(x-y) == 1: continue
        if x > y: z = x; x = y ; y = z
        tb.remove( x )
        tb.remove( y )
        if STECKER_NUMB:
            if ZERO:
                steckerverbindungen += "%02d/%02d " % (x+1,y+1)
            else:
                if (y+1) < 10:
                    steckerverbindungen += "%2d/%1d  " % (x+1,y+1)
                else:
                    steckerverbindungen += "%2d/%2d " % (x+1,y+1)
        else:
            steckerverbindungen += ALPHA[x] + ALPHA[y] + " "
        nbplugs += 1
    if MAX_PLUGS != MIN_PLUGS:
        diff = NB_OF_PLUGS - nb_of_plugs
        if STECKER_NUMB:
            steckerverbindungen += "      " * diff
        else:
            steckerverbindungen += "   " * diff
    return steckerverbindungen
# ====================================================
def kenn():
    kenngruppen = ""
    for j in range(4):
        kenngruppen += allKenn.pop().lower() + " "
    return kenngruppen
# ====================================================
def usage():
    print("Usage:")
    print(" -h          This help")
    print(" -N nets     The number of Networks (by default one)")
    print(" -H header   Header, you can use this several times")
    print("             If nets > 1, the headers will be different")
    print(" -n          No Fields (remove the name of fields)")
    print(" -W n        The nomber of rotors [3,5,8], 5 by default")
    print(" -R          Ringstellung in letter (in number by default)")
    print(" -G          Grundstellung in letter (in number by default)")
    print(" -K          Print the Kennzeichnung in place of Kenngruppen")
    print(" -D          Print the key only for one day")
    print(" -M          Print a whole month (only one key by default)")
    print(" -S          Steckers in number (in letter by default)")
    print(" -P n        The numbers of steckers (10 by default)")
    print("             If a comma is present, that means -P min,max")
    print(" -C          The CSKO constraint applies")
    print(" -F fields   The fields, by default:  -F DWRSK")
    print("    D: Day, W: Walzenlage, R: Ringstellung, G: Grundstellung")
    print("    S: Steckers, K: Kenngruppen, E: Space, e: Space[in header]")
    print("    navy: Day(only one two), Walz above Ring")
    print(" -Z          Print numbers preceded by zero")
# ====================================================
# Manage options
try:
    opts, args = getopt.getopt(sys.argv[1:],
        "hH:N:nW:RGKDMSP:CF:Z", ["help", "header=", "nbofnets=", "nofields", 
        "numbofwalz=",  "ringinlet", "grundinlet", "kennzeichnung", "aday",
        "awholemonth", "steckersinnumb", "numbofplugs=", "csko", "format=", "ZERO"  ])
except getopt.GetoptError as err:
    usage()
    sys.exit(2)
NB_OF_NETS = 1
NB_OF_PLUGS = 10
MAX_PLUGS = 10
MIN_PLUGS = 10
NB_OF_WALZS = 5
RING_NUMB = True
GRUND_NUMB = True
KENNZEICHNUNG = False
CSKO = False
MONTH = False
ADAY = False
STECKER_NUMB = False
FORMAT_MONTH = "WreeeeESeeeeeeeEK"
FORMAT_DAY = "WRGSK"
FORMAT = ""
HEADER = ""
NOFIELDS = False
HEAD_NETS = []
HEADS_nb = 0
ZERO = False

for o, a in opts:
    if o == "-v":
        DEBUG = True
    elif o in ("-h", "--help"):
        usage()
        sys.exit()
    elif o in ("-H", "--header"):
        if NB_OF_NETS > 1 and HEADS_nb < NB_OF_NETS:
            HEAD_NETS.append( a ) ; HEADS_nb += 1
        else:
            HEADER +=  a + "\n"
    elif o in ("-N", "--nbofnets"):
        NB_OF_NETS = int(a)
    elif o in ("-n", "--nofields"):
        NOFIELDS = True
    elif o in ("-W", "--numbofwalz"):
        NB_OF_WALZS = int(a)
    elif o in ("-R", "--ringinlet"):
        RING_NUMB = False
    elif o in ("-G", "--grundinlet"):
        GRUND_NUMB = False
    elif o in ("-K", "--kennzeichnung"):
        KENNZEICHNUNG = True
        FIELDS['K'] = FIELDS['k']
    elif o in ("-D", "--aday"):
        ADAY = True
    elif o in ("-M", "--awholemonth"):
        MONTH = True
    elif o in ("-S", "--steckersinnumb"):
        STECKER_NUMB = True
    elif o in ("-P", "--numbofplugs"):
        if "," in a:
            x,y = a.split(",")
            NB_OF_PLUGS = MAX_PLUGS = int(y)
            MIN_PLUGS = int(x)
        else:
            NB_OF_PLUGS = int(a)
    elif o in ("-C", "--csko"):
        CSKO = True
    elif o in ("-F", "--format"):
        FORMAT = a
    elif o in ("-Z", "--zero"):
        ZERO = True
    else:
        assert False, "unhandled option"
if not ADAY and not MONTH:
    usage()
    sys.exit(1)
# ====================================================
# ----------- Start of the program -------------------
allKenn = []
genAllKenn()
random.shuffle( allKenn )

allKzeich = []
genKzeich()

if FORMAT == "":
    if MONTH: FORMAT = FORMAT_MONTH
    else: FORMAT = FORMAT_DAY
if FORMAT == "navy":
    if HEADER != "": print(HEADER.rstrip())
    RING_NUMB = False
    print("Tag Innere Einstellung")
    for i in range(31,0,-2):
        print("%3d " % (i), walz(8))
        ch = ring()
        print("     %-4s %-4s %-4s" % (ch[0],ch[2],ch[4]))
    sys.exit(0)
if ADAY:
    if HEADER != "": print(HEADER)
    for aField in FORMAT:
        if aField == 'W':
            print( "%-24s : %s" % (FIELDS[aField], walz( NB_OF_WALZS) ) )
        if aField == 'R':
            print( "%-24s : %s" % (FIELDS[aField], ring() ))
        if aField == 'G':
            print( "%-24s : %s" % (FIELDS[aField], grund() ))
        if aField == 'S':
            print( "%-24s : %s" % (FIELDS[aField], steckers() ))
        if aField == 'K':
            print( "%-24s : %s" % (FIELDS[aField], kenn() ))
    sys.exit(0)
if MONTH:
    for aNet in range(NB_OF_NETS):    
        if HEADS_nb > 0 : print( HEAD_NETS[ aNet ] )
        if HEADER != "": print(HEADER.rstrip())
        if not NOFIELDS:
            for aField in FORMAT:
                ligne = "Tag "
                for aField in FORMAT:
                    if aField == 'E': continue
                    ligne += FIELDS[ aField ]    
            print(ligne)
        for i in range(31):
            ligne = "%3d " % (31-i)
            for aField in FORMAT:
                if aField == 'e': continue
                if aField == 'W':
                    ligne += walz( NB_OF_WALZS) 
                if aField.upper() == 'R':
                    ligne += ring() 
                if aField.upper() == 'G':
                    ligne += grund() 
                if aField == 'S':
                    ligne += steckers() 
                if aField == 'K':
                    if KENNZEICHNUNG:
                        ligne += allKzeich[i][aNet]
                    else:
                        ligne += kenn() 
                if aField == 'E':
                   ligne += ' '
            print(ligne)
